%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.0, (the "License"); you may not use this file except in
%% compliance with the License. You may obtain a copy of the License at
%% http://www.erlang.org/EPL1_0.txt
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Original Code is Erlang-4.7.3, December, 1998.
%% 
%% The Initial Developer of the Original Code is Ericsson Telecom
%% AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
%% Telecom AB. All Rights Reserved.
%% 
%% Contributor(s): ______________________________________.''
%%
-module(jive).
-copyright('Copyright (c) 1991-97 Ericsson Telecom AB').
-author('kaj@medialab.ericsson.se').
-export([start/0,start/1,stop/0,get_pid/1,register_pid/1,unregister_pid/1,
	 list_to_string/1,string_to_list/1,allow/1,listen_boot/1,
	 jive_server_boot/0,listen/1,pack/1]). 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Exported functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% Spawn the server at port 14000
%%
start() ->
  start(14000).

%%    
%% Spawn the server at port Port
%%
start(Port) ->
  case whereis(jive_server) of
    undefined ->
      Pid = spawn(jive,jive_server_boot,[]),
      register(jive_server,Pid),
      proc_lib:spawn(?MODULE, listen_boot, [Port]),
      Pid;
    Pid ->
      Pid
  end.

%%    
%% Stop the server
%%
stop() ->
  jive_server ! shutdown.

%%
%% Get the pid from a pid-id
%%
get_pid(PidId) ->
  jive_server ! {get,PidId,self()},
  receive 
    {value,{_,Pid}} when pid(Pid) -> 
      Pid;
    false ->
      error
  end.

%%
%% register a pid
%%
register_pid(Pid) when pid(Pid) ->
  jive_server ! {set,Pid,self()},
  receive 
    {pidid,PidId} when integer(PidId) ->
      {pidid,PidId};
    _ ->
      %% should never happen
      error  
  end.

%%
%% unregister a pid
%%
unregister_pid(Pid) when pid(Pid) ->
  jive_server ! {remove,Pid}.

%%
%% allow Java to connect to a function in a module
%%
allow({Module,Function, Arity}) ->
  jive_server ! {allow,{Module,Function, Arity}};

%%
%% allow Java to connect to all functions in all available modules (for debugging)
%%
allow(all) ->
  jive_server ! {allow,all}.

%%
%% Convert a list to a string, no checking is done
%%
list_to_string(List) ->
  {string, List}.

%%
%% Convert a string to a list, no checking is done
%%
string_to_list({string, List}) ->
  List.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Jive server
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

jive_server_boot() ->
  process_flag(trap_exit,true),
  jive_server_loop([],[],0).

jive_server_loop(Allowed, Pids, PidIdNr) -> 
  receive
    {set,Pid,Sender} when pid(Pid) ->
      link(Pid),
      PidId = {pidid,PidIdNr},
      PidTuple = {PidId , Pid},
      NPids = [PidTuple|Pids],
      Sender ! PidId,
      jive_server_loop(Allowed,NPids, PidIdNr+1);

    {remove,Pid} when pid(Pid) ->
      unlink(Pid),
      NPids = lists:keydelete(Pid,2,Pids),
      jive_server_loop(Allowed,NPids, PidIdNr);

    {get,PidId,Sender}  ->
      Tuple = lists:keysearch(PidId,1,Pids),
      Sender ! Tuple,
      jive_server_loop(Allowed, Pids,PidIdNr);

    {allow,all} ->
      jive_server_loop(all, Pids, PidIdNr);
    
    {allow,Tuple} when tuple(Tuple) ->
      case Allowed of
	all ->
	  jive_server_loop(Allowed,Pids,PidIdNr);
	AList when list(AList) ->
	  jive_server_loop([Tuple|AList],Pids,PidIdNr)
      end;

    {check_allowed, Tuple, ReturnPid} ->
      case Allowed of	
	all ->
	  ReturnPid ! ok;
	AList when list(AList) ->
	  case lists:member(Tuple, AList) of 
	    true ->
	      ReturnPid ! ok;
	    false ->
	      ReturnPid ! error
	  end
      end,
      jive_server_loop(Allowed,Pids,PidIdNr);

    {'EXIT',FromPid,Reason} ->
      jive_server_loop(Allowed,Pids,PidIdNr);    
    
    shutdown ->
      shutdown
  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Jive Socket listener
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

listen_boot(Port) ->
    ListenSocket = tcp_listen(Port),
    listen(ListenSocket).

listen(ListenSocket) ->
    Accepted = tcp_accept(ListenSocket),
    proc_lib:spawn(jive,listen,[ListenSocket]),    
    listen_loop(Accepted).

listen_loop(Socket) ->
    receive
	{tcp_closed, Socket} ->
	    server_done;

	{tcp_closed, Socket, Message} ->	% FIXME: Actually an error report
	    server_done;

	{tcp, Socket, Data} ->
	    case parse_message(Data) of 
		{new_client, PidId} ->
		    tcp_send(Socket, [pack(PidId)]),
		    jive_client_boot(Socket);
		{value,Val} ->
		    tcp_send(Socket, [pack(Val)]);
		{error, ErrNo, Msg} ->
		    tcp_send(Socket, [60,ErrNo,pack(Msg)]);
		error ->			% FIXME: This is meaningless
		    error;
		ok ->				% FIXME: This is meaningless
		    ok
	    end,
	    listen_loop(Socket)

	% FIXME: We wait forever......
  end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Java Client Listener
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

jive_client_boot(Socket) ->
  jive_client_loop(Socket).

jive_client_loop(Socket) -> 
  receive

    %% Client disconnected Socket is closed. This listen process
    %% unregister itself from the jive server

    {tcp_closed, Socket} ->
      unregister_pid(self()),
      exit(normal);

    {tcp_closed, Socket, Data} ->		% FIXME: Close because of error
      unregister_pid(self()),
      exit(normal);

    %% We got data so we send it thru the socket to the client

    {send,Receiver,Msg} ->
      SendMsg  = pack_message(Receiver,Msg),
      tcp_send(Socket, [pack_message(Receiver,Msg)])
  end,
  jive_client_loop(Socket).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Parse the different messages that can be sent to the server
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% Parse a message sent to the server
%%
parse_message(Msg) ->
  [Tag|Tail] = Msg,
  case Tag of 
    50 ->
      parse_apply(Msg);
    51 ->
      parse_spawn(Msg);
    52 ->
      parse_send(Msg);
    53 ->
      parse_new_client(Msg)
  end.

%%
%% Parse an apply message
%%
parse_apply([50 | Message]) ->
  {Module,Msg2} = parse_atom(Message),
  {Name,Msg3} = parse_atom(Msg2),
  {List,Null} = parse_list(Msg3),
  case check_allowed(Module,Name, length(List)) of 
    ok ->
      Result = apply(Module,Name,List),
      {value,Result};
    error ->
      ErrMsg = lists:concat(["Function ",Module,":",Name,"/",
			     length(List),
			     " is not allowed to be used"]),
      do_error(1,ErrMsg)
  end.

%%
%% Parse a spawn message
%%
parse_spawn([51 | Message]) ->
  {Module,Msg2} = parse_atom(Message),
  {Name,Msg3} = parse_atom(Msg2),
  {List,Null} = parse_list(Msg3),
  case check_allowed(Module,Name, length(List)) of 
    ok ->
      Pid = proc_lib:spawn(Module,Name,List),
      PidId = register_pid(Pid),
      {value,PidId};
    error ->
      ErrMsg = lists:concat(["Process ",Module,":",Name,"/",length(List),
			     " is not allowed to be spawned"]),
      do_error(1,ErrMsg)
  end.

%%
%% Parse a message sent when a client connects
%%
parse_new_client([53 | Message]) ->
  PidId = register_pid(self()),
  {new_client,PidId}.

%%
%% Parse a send message to a pid
%%
parse_send([52 | Message]) ->
  {PidId,Expr} = parse_pid(Message),
  {Msg,Null} = parse_expr(Expr), 
  Val = get_pid(PidId),
  case Val of 
    error ->
      error;
    Pid when pid(Pid) ->
      Pid ! Msg,
      ok
  end.

%%
%% Check if Java client is allowed to use this function and module
%%
check_allowed(Module, Function, Arity) ->
  jive_server ! {check_allowed, {Module,Function, Arity}, self()},
  receive
    Val when atom(Val) ->
      Val
  end.

do_error(ErrNo, Message) when integer(ErrNo), list(Message) ->
  {error, ErrNo, list_to_string(Message)}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Parse an encoded Erlang expression
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% Parse an encoded Erlang expression
%%
parse_expr(Expr) ->
  [Tag|Tail]  = Expr,
  case Tag of 
    100 ->
      parse_string(Expr);
    101 ->
      parse_list(Expr);
    102 ->
      parse_tuple(Expr);
    103 ->
      parse_atom(Expr);
    104 ->
      parse_integer(Expr);
    105 ->
      parse_pid(Expr);
    106 ->
      parse_float(Expr);
    107 ->
      parse_binary(Expr)
  end.

%%
%% Parse an encoded string
%%
parse_string([100,L1,L2,L3,L4|Message]) ->
  Length = bytes_to_int(L1,L2,L3,L4),
  String = lists:sublist(Message,1,Length),
  Tail = lists:nthtail(Length,Message),
  {String,Tail}.

%%
%% Parse an encoded atom
%%
parse_atom([103,L1,L2,L3,L4|Message]) ->
  Length = bytes_to_int(L1,L2,L3,L4),
  String = lists:sublist(Message,1,Length),
  Tail = lists:nthtail(Length,Message),
  Atom = list_to_atom(String),
  {Atom,Tail}.

%%
%% Parse an encoded integer (sign + unsigned int)
%%
parse_integer([104|Message]) ->
  [Sign, I1,I2,I3,I4|Tail]=Message,
  Int = bytes_to_int(I1,I2,I3,I4),
  case Sign of 
    0 ->
      {Int,Tail};
    1 ->
      {-Int,Tail}
  end.

%%
%% Parse an encoded pid-id
%%
parse_pid([105|Message]) ->
  [I1,I2,I3,I4|Tail]=Message,
  Pid = bytes_to_int(I1,I2,I3,I4),
  {{pidid,Pid},Tail}.

%%
%% Parse an encoded list
%%
parse_list([101,L1,L2,L3,L4 | Message]) ->
  Length = bytes_to_int(L1,L2,L3,L4),
  parse_list(Message,Length,[]).
parse_list(Tail,0,List) ->
  RList = lists:reverse(List),
  {RList,Tail};
parse_list(Message,Length,List) ->
  {Expr,Tail} = parse_expr(Message),
  parse_list(Tail,Length-1,[Expr | List]).

%%
%% Parse an encoded tuple
%%
parse_tuple([102,L1,L2,L3,L4|Message]) ->
  Length = bytes_to_int(L1,L2,L3,L4),
  {List,Tail} = parse_list(Message,Length,[]),
  Tuple = list_to_tuple(List),
  {Tuple,Tail}.

%%
%% Parse an encoded float
%%
parse_float([106|Message]) ->
  {Float,Tail} = parse_string([100|Message]),
  {list_to_float(Float),Tail}.

%%
%% Parse an encoded binary
%%
parse_binary([107|Message]) ->
  {Binary,Tail} = parse_string([100|Message]),
  {list_to_binary(Binary),Tail}.

%%
%% convert 4 bytes to an unsigned int
%%
bytes_to_int(I1,I2,I3,I4) ->
  (I1 bsl 24) bor (I2 bsl 16) bor (I3 bsl 8) bor I4.

%%
%% convert an int to 4 bytes
%%
int_to_bytes(I) ->
  [((I band 16#ff000000) bsr 24),((I band 16#ff0000) bsr 16),
   ((I band 16#ff00) bsr 8),I band 16#ff].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Pack messages sent to the java side
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% Packs a message to a receiver
%%
pack_message(Receiver,Message) when integer(Receiver) ->
  lists:concat([pack(Receiver),pack(Message)]).

%%
%% Pack a string
%%
pack({string,S}) when list(S) ->
  lists:append([100|int_to_bytes(length(S))],S);


%%
%% Pack a PidID
%%
pack({pidid,I}) when integer(I) ->
  [105|int_to_bytes(I)];

%%
%% Pack a list
%%
pack(L) when list(L) ->
  Header = [101|int_to_bytes(length(L))],
  ByteList = lists:map({jive,pack},[],L),
  lists:concat([Header|ByteList]);

%%
%% Pack a tuple
%%
pack(T) when tuple(T) ->
  L = tuple_to_list(T),
  Header = [102|int_to_bytes(length(L))],
  ByteList = lists:map({jive,pack},[],L),
  lists:concat([Header|ByteList]);

%%
%% Pack an atom
%%
pack(A) when atom(A) ->
  S = atom_to_list(A),
  lists:append([103,int_to_bytes(length(S))],S);

%%
%% Pack an Integer
%%
pack(I) when integer(I) ->
  [104|int_to_bytes(I)];

%%
%% Pack a float
%%
pack(F) when float(F) ->
  S = float_to_list(F),
  lists:append([106|int_to_bytes(length(S))],S);

%%
%% Pack a binary
%%
pack(B) when binary(B) ->
  L = binary_to_list(B),
  lists:append([107|int_to_bytes(length(L))],L).

%%
%% Low level stuff
%%
tcp_listen(PortNo) ->
    {ok, ListenSocket} = gen_tcp:listen(PortNo, [{packet,4}, {reuseaddr, true}]),
    ListenSocket.

tcp_accept(ListenSocket) ->
    {ok, Sock} = gen_tcp:accept(ListenSocket),
    Sock.

tcp_send(Socket, Msg) ->
    gen_tcp:send(Socket, Msg).
